/*******************************************************************************

                              MPJ2 Block cipher

                         implemented by A.Myasnikow

*******************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <mem.h>

typedef int boolean;
typedef unsigned char u8;
typedef unsigned long int u32;

static u8 *key = NULL;
static u32 keysize;
static u32 keyindex;
static unsigned long accum;
static u32 numrounds;
static u8 s[61440];           /* Substitution boxes. */
static u8 si[61440];          /* Inverse substitution boxes. */

#define CRC_MASK           0xFFFFFFFFL
#define CRC32_POLYNOMIAL   0xEDB88320L


unsigned long int Ccitt32Table[256];

/****************************************************************************/

/*
 * This routine simply builds the coefficient table used to calculate
 * 32 bit CRC values throughout this program.  The 256 long word table
 * has to be set up once when the program starts.  Alternatively, the
 * values could be hard coded in, which would offer a miniscule improvement
 * in overall performance of the program.
 */

int
BuildCRCTable (void)
{
  int i;
  int j;
  unsigned long value;


  for (i = 0; i !=256; i++)
    {
      value = i;
      for (j = 8; j > 0; j--)
        {
          if (value & 1)
            value = (value >> 1) ^ CRC32_POLYNOMIAL;
          else
            value >>= 1;
        }
      Ccitt32Table[i] = value;
    }
  return 0;
}

#define crc32(crc, c) (((crc >> 8) & 0x00FFFFFFL) ^ (Ccitt32Table[(int)((int) crc ^ c) & 0xFF]))

#define false (0)
#define true (1)


u32
keyrand (u32 max_value)        // Returns uniformly distributed pseudorandom
{                               // value based on key[], sized keysize
  u32 prandvalue, i;
  unsigned long mask;

  if (!max_value)
    return 0;
    
  mask = 0L;                    // Create a mask to get the minimum
  for (i = max_value; i > 0; i = i >> 1)        // number of bits to cover the
    mask = (mask << 1) | 1L;    // range 0 to max_value.
  i = 0;
  do
    {
      accum = crc32 (accum, key[keyindex++]);
      if (keyindex >= keysize)
        keyindex = 0;           // Recycle thru the key
      prandvalue = (u32) (accum & mask);
      if ((++i > 97) && (prandvalue > max_value))       // Don't loop forever.
        prandvalue -= max_value;        // Introduce negligible bias.
    }
  while (prandvalue > max_value);       // Discard out of range values.
  return prandvalue;
}

static void
makeonebox (u32 i, u32 j)
{
  int n;
  u32 pos, m, p;
  boolean filled[256];

  for (m = 0; m < 256; m++)     /* The filled array is used to make sure that */
    filled[m] = false;          /* each u8 of the array is filled only once. */
  for (n = 255; n >= 0; n--)    /* n counts the number of u8s left to fill */
    {
      pos = keyrand (n);        /* pos is the position among the UNFILLED */
      /* components of the s array that the */
      /* number n should be placed.  */
      p = 0;
      while (filled[p])
        p++;
      for (m = 0; m < pos; m++)
        {
          p++;
          while (filled[p])
            p++;
        }
      *(s + (4096 * i) + (256 * j) + p) = n;
      filled[p] = true;
    }
}

void __stdcall __export
setup (u8 * external_key)
/* This procedure generates internal keys by filling the substitution box array
  s based on the external key given as input.  It DOES take a bit of time. */
{
  u32 i, j, k;

  numrounds = 15;

  BuildCRCTable ();


  key = external_key;
  keysize = 16;

  accum = 0xFFFFFFFFL;

  for (i = 0; i < numrounds; i++)
    {
      for (j = 0; j < 16; j++)
        {
          makeonebox (i, j);
        }
    }


  {
 

    for (i = 0; i < numrounds; i++)
      {
        for (j = 0; j < 16; j++)
          {
            for (k = 0; k < 256; k++)
              {
                *(si + (4096 * i) + (256 * j) +
                  *(s + (4096 * i) + (256 * j) + k)) = k;
              }
          }
      }
  }
}

static void
permute (u8 * x, u8 * y)    // x and y must be different.
/* This procedure is designed to make each bit of the output dependent on as
  many u8s of the input as possible, especially after repeated application.
  Each output u8 takes its least significant bit from the corresponding
  input u8.  The next higher bit comes from the corresponding bit of the
  input u8 to the left.  This is done until all bits of the output u8
  are filled.  Where there is no u8 to the left, the u8 at the far right
  is used. */
{
  y[0] = (x[0] & 1) | (x[1] & 2) | (x[2] & 4) |
    (x[3] & 8) | (x[4] & 16) | (x[5] & 32) | (x[6] & 64) | (x[7] & 128);
  y[1] = (x[1] & 1) | (x[2] & 2) | (x[3] & 4) |
    (x[4] & 8) | (x[5] & 16) | (x[6] & 32) | (x[7] & 64) | (x[8] & 128);
  y[2] = (x[2] & 1) | (x[3] & 2) | (x[4] & 4) |
    (x[5] & 8) | (x[6] & 16) | (x[7] & 32) | (x[8] & 64) | (x[9] & 128);
  y[3] = (x[3] & 1) | (x[4] & 2) | (x[5] & 4) |
    (x[6] & 8) | (x[7] & 16) | (x[8] & 32) | (x[9] & 64) | (x[10] & 128);
  y[4] = (x[4] & 1) | (x[5] & 2) | (x[6] & 4) |
    (x[7] & 8) | (x[8] & 16) | (x[9] & 32) | (x[10] & 64) | (x[11] & 128);
  y[5] = (x[5] & 1) | (x[6] & 2) | (x[7] & 4) |
    (x[8] & 8) | (x[9] & 16) | (x[10] & 32) | (x[11] & 64) | (x[12] & 128);
  y[6] = (x[6] & 1) | (x[7] & 2) | (x[8] & 4) |
    (x[9] & 8) | (x[10] & 16) | (x[11] & 32) | (x[12] & 64) | (x[13] & 128);
  y[7] = (x[7] & 1) | (x[8] & 2) | (x[9] & 4) |
    (x[10] & 8) | (x[11] & 16) | (x[12] & 32) | (x[13] & 64) | (x[14] & 128);
  y[8] = (x[8] & 1) | (x[9] & 2) | (x[10] & 4) |
    (x[11] & 8) | (x[12] & 16) | (x[13] & 32) | (x[14] & 64) | (x[15] & 128);
  y[9] = (x[9] & 1) | (x[10] & 2) | (x[11] & 4) |
    (x[12] & 8) | (x[13] & 16) | (x[14] & 32) | (x[15] & 64) | (x[0] & 128);
  y[10] = (x[10] & 1) | (x[11] & 2) | (x[12] & 4) |
    (x[13] & 8) | (x[14] & 16) | (x[15] & 32) | (x[0] & 64) | (x[1] & 128);
  y[11] = (x[11] & 1) | (x[12] & 2) | (x[13] & 4) |
    (x[14] & 8) | (x[15] & 16) | (x[0] & 32) | (x[1] & 64) | (x[2] & 128);
  y[12] = (x[12] & 1) | (x[13] & 2) | (x[14] & 4) |
    (x[15] & 8) | (x[0] & 16) | (x[1] & 32) | (x[2] & 64) | (x[3] & 128);
  y[13] = (x[13] & 1) | (x[14] & 2) | (x[15] & 4) |
    (x[0] & 8) | (x[1] & 16) | (x[2] & 32) | (x[3] & 64) | (x[4] & 128);
  y[14] = (x[14] & 1) | (x[15] & 2) | (x[0] & 4) |
    (x[1] & 8) | (x[2] & 16) | (x[3] & 32) | (x[4] & 64) | (x[5] & 128);
  y[15] = (x[15] & 1) | (x[0] & 2) | (x[1] & 4) |
    (x[2] & 8) | (x[3] & 16) | (x[4] & 32) | (x[5] & 64) | (x[6] & 128);
}

static void
ipermute (u8 * x, u8 * y)   /* x!=y */
/* This is the inverse of the procedure permute. */
{
  y[0] = (x[0] & 1) | (x[15] & 2) | (x[14] & 4) |
    (x[13] & 8) | (x[12] & 16) | (x[11] & 32) | (x[10] & 64) | (x[9] & 128);
  y[1] = (x[1] & 1) | (x[0] & 2) | (x[15] & 4) |
    (x[14] & 8) | (x[13] & 16) | (x[12] & 32) | (x[11] & 64) | (x[10] & 128);
  y[2] = (x[2] & 1) | (x[1] & 2) | (x[0] & 4) |
    (x[15] & 8) | (x[14] & 16) | (x[13] & 32) | (x[12] & 64) | (x[11] & 128);
  y[3] = (x[3] & 1) | (x[2] & 2) | (x[1] & 4) |
    (x[0] & 8) | (x[15] & 16) | (x[14] & 32) | (x[13] & 64) | (x[12] & 128);
  y[4] = (x[4] & 1) | (x[3] & 2) | (x[2] & 4) |
    (x[1] & 8) | (x[0] & 16) | (x[15] & 32) | (x[14] & 64) | (x[13] & 128);
  y[5] = (x[5] & 1) | (x[4] & 2) | (x[3] & 4) |
    (x[2] & 8) | (x[1] & 16) | (x[0] & 32) | (x[15] & 64) | (x[14] & 128);
  y[6] = (x[6] & 1) | (x[5] & 2) | (x[4] & 4) |
    (x[3] & 8) | (x[2] & 16) | (x[1] & 32) | (x[0] & 64) | (x[15] & 128);
  y[7] = (x[7] & 1) | (x[6] & 2) | (x[5] & 4) |
    (x[4] & 8) | (x[3] & 16) | (x[2] & 32) | (x[1] & 64) | (x[0] & 128);
  y[8] = (x[8] & 1) | (x[7] & 2) | (x[6] & 4) |
    (x[5] & 8) | (x[4] & 16) | (x[3] & 32) | (x[2] & 64) | (x[1] & 128);
  y[9] = (x[9] & 1) | (x[8] & 2) | (x[7] & 4) |
    (x[6] & 8) | (x[5] & 16) | (x[4] & 32) | (x[3] & 64) | (x[2] & 128);
  y[10] = (x[10] & 1) | (x[9] & 2) | (x[8] & 4) |
    (x[7] & 8) | (x[6] & 16) | (x[5] & 32) | (x[4] & 64) | (x[3] & 128);
  y[11] = (x[11] & 1) | (x[10] & 2) | (x[9] & 4) |
    (x[8] & 8) | (x[7] & 16) | (x[6] & 32) | (x[5] & 64) | (x[4] & 128);
  y[12] = (x[12] & 1) | (x[11] & 2) | (x[10] & 4) |
    (x[9] & 8) | (x[8] & 16) | (x[7] & 32) | (x[6] & 64) | (x[5] & 128);
  y[13] = (x[13] & 1) | (x[12] & 2) | (x[11] & 4) |
    (x[10] & 8) | (x[9] & 16) | (x[8] & 32) | (x[7] & 64) | (x[6] & 128);
  y[14] = (x[14] & 1) | (x[13] & 2) | (x[12] & 4) |
    (x[11] & 8) | (x[10] & 16) | (x[9] & 32) | (x[8] & 64) | (x[7] & 128);
  y[15] = (x[15] & 1) | (x[14] & 2) | (x[13] & 4) |
    (x[12] & 8) | (x[11] & 16) | (x[10] & 32) | (x[9] & 64) | (x[8] & 128);
}

static void
substitute (u32 round, u8 * x, u8 * y)
{
  u32 i;

  for (i = 0; i < 16; i++)
    y[i] = *(s + (4096 * round) + (256 * i) + x[i]);
}

static void
isubst (u32 round, u8 * x, u8 * y)
{
  u32 i;

  for (i = 0; i < 16; i++)
    y[i] = *(si + (4096 * round) + (256 * i) + x[i]);
}


void __stdcall __export
crypt (u8 * x)

{
  u32 round, i;
  u8 z[16], y[16];


  substitute (0, x, y);
  for (round = 1; round < numrounds; round++)
    {
      permute (y, z);
      substitute (round, z, y);
    }


  for (i = 0; i < 16; i++)
    {
      x[i] = y[i];
    }


}

void __stdcall __export
decrypt (u8 * x)

{
  int round, i;
  u8 z[16], y[16];

  isubst (numrounds - 1, x, y);
  for (round = numrounds - 2; round >= 0; round--)
    {
      ipermute (y, z);
      isubst (round, z, y);
    }

  for (i = 0; i < 16; i++)
    {
      x[i] = y[i];
    }

}




#include <string.h>


typedef unsigned char u8;
typedef unsigned long u32;

u32 __stdcall __export
getblocksize ()
{
  return 128;
}

u32 __stdcall __export
getkeysize ()
{
  return 128;
}

void __stdcall __export
getciphername (u8 * p)
{
  strcpy (p, "MPJ2");
}
